/***
 * Juliac
 * Copyright (C) 2008-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.adl;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.adl.stati.ComponentFactoryClassGenerator;
import org.objectweb.fractal.juliac.api.ADLParserSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.desc.ComponentDescFactoryClassGenerator;

/**
 * This class implements the services provided by Fractal ADL to Juliac.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.0
 */
public class FractalADLSupportImpl implements ADLParserSupportItf {

    /**
     * The Fractal ADL factory component.
     */
    final static public String FACTORY =
        "org.objectweb.fractal.juliac.adl.JuliacFactory";
    
    /**
     * The Fractal ADL backend component for retrieving the model of an ADL
     * descriptor.
     */
    final static public String MODEL_BACKEND =
        "org.objectweb.fractal.juliac.adl.JuliacModelBackend";

    /**
     * The Fractal ADL backend component which generates the source code of the
     * {@link org.objectweb.fractal.api.factory.Factory} class for an ADL
     * descriptor.
     */
    final static public String STATIC_BACKEND =
        "org.objectweb.fractal.juliac.adl.JuliacStaticBackend";

    
    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	final public static Class<ADLParserSupportItf> SERVICE_TYPE =
		ADLParserSupportItf.class;
	
	protected Juliac jc;
	
    /**
     * Initialize the service.
     */
    public void init( Juliac jc ) {
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
    }
    
    public void close( Juliac jc ) {
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }

    
    // -----------------------------------------------------------------------
    // Implementation of the ADLParserSupportItf interface
    // -----------------------------------------------------------------------

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified ADL descriptor.
     * 
     * @param adl  the ADL descriptor to be checked
     * @return     if the ADL descriptor should be accepted
     * @since 2.2.1
     */
    public boolean acceptADLDesc( String adl ) {
        /*
         * Do not use File#separatorChar (even on Windows) as it leads to is
         * being null when the resource is stored in a jar file.
         */
        String adlFileName = adl.replace('.','/')+".fractal";
        JuliacConfig jconf = jc.getJuliacConfig();
        ClassLoader loader = jconf.getClassLoader();
        InputStream is = loader.getResourceAsStream(adlFileName);
        boolean exist = (is != null);
        return exist;
    }
    
    /**
     * Parse the specified ADL and return a model of it.
     * 
     * @param adl      the fully-qualified name of the ADL descriptor
     * @param context  contextual information
     * @return         the model of the specified ADL
     */
    public ComponentDesc parse( String adl, Map<Object,Object> context )
    throws IOException {

    	Map<Object,Object> newcontext = getContext(context);
    	
    	Object ret = null; 
        
    	try {
            Factory juliacADLFactory =
                FactoryFactory.getFactory(
                    FractalADLSupportImpl.FACTORY,
                    FractalADLSupportImpl.MODEL_BACKEND,
                    new HashMap<Object,Object>() );
            ret = juliacADLFactory.newComponent(adl,newcontext);
        }
        catch( ADLException ae ) {
            /*
             * Don't use new IOException(ae) as IOException#<init>(Throwable)
             * was introduced in JDK 6 and will generate an error when compiling
             * with JDK 5.
             */
            IOException ioe = new IOException();
            ioe.initCause(ae);
            throw ioe;
        }
        
        if( !(ret instanceof ComponentDesc) ) {
            String msg =
                "Return value expected to be of type ComponentDesc. Got: "+
        		ret.getClass().getName();
            throw new IOException(msg);
        }
        
        return (ComponentDesc) ret;
    }

    /**
     * Generate the source code of a factory class for the specified ADL.
     * 
     * @param adl         the fully-qualified name of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     */
    public void generate( String adl, String targetname )
    throws IOException {

    	Map<Object,Object> context = getContext(null);
    	
    	Factory juliacADLFactory = null;
    	try {
            juliacADLFactory =
                FactoryFactory.getFactory(
                    FractalADLSupportImpl.FACTORY,
                    FractalADLSupportImpl.STATIC_BACKEND,
                    new HashMap<Object,Object>() );
        }
        catch( ADLException ae ) {
            /*
             * Don't use new IOException(ae) as IOException#<init>(Throwable)
             * was introduced in JDK 6 and will generate an error when compiling
             * with JDK 5.
             */
            IOException ioe = new IOException();
            ioe.initCause(ae);
            throw ioe;
        }
    	
		ComponentFactoryClassGenerator cfcg = 
            new ComponentFactoryClassGenerator(
        		adl, targetname, juliacADLFactory, context );
        jc.generateSourceCode(cfcg);
    }

    /**
     * Generate the source code of a factory class for the specified model of
     * the ADL.
     * 
     * @param cdesc       the model of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     * @since 2.5
     */
    public void generate( ComponentDesc cdesc, String targetname )
    throws IOException {
    	
		ComponentDescFactoryClassGenerator cfcg = 
            new ComponentDescFactoryClassGenerator(jc,cdesc,targetname);
        jc.generateSourceCode(cfcg);
    }

    
    // -----------------------------------------------------------------------
    // Implementation specific
    // -----------------------------------------------------------------------

    protected Map<Object,Object> getContext( Map<Object,Object> context ) {
    	
    	Map<Object,Object> contextMap =
    		context==null ? new HashMap<Object,Object>(): context;
        	
        /*
         * Initialize a context which includes all system properties.
         * This has been added following a discussion with Alessio who wanted to
         * be able to transmit argument values to compile with Juliac the code
         * of Fractal ADL.
         * This is also useful for any ADL which is parameterized.
         */
        Map<Object,Object> newcontext = new HashMap<Object,Object>();
        newcontext.putAll(contextMap);
        
        Map<Object,Object> systemPropertiesMap =
            new HashMap<Object,Object>(System.getProperties());
        newcontext.putAll(systemPropertiesMap);
                
        /*
         * Instantiates YesClassLoader with the Juliac classloader.
         * If this is not done (e.g. just like instantiating
         * YesClassLoader without any parent), a default class loader is used.
         * When run under maven, this default class loader cannot load
         * resource streams such as the .fractal files which are bundled in
         * dependent .jar files (such as koch-runtime.jar).
         * The trick is not needed under eclipse, but the code is kept as this
         * to preserve compatibility with maven.
         */
        ClassLoader classLoader = jc.getJuliacConfig().getClassLoader();
        ClassLoader yesClassLoader = new YesClassLoader(classLoader);
        newcontext.put("classloader",yesClassLoader);
        
        newcontext.put(CONTEXT_ID_JULIAC,jc);
        
        return newcontext;
    }
    
    /**
     * The key associated to the instance of Juliac in the context map used by
     * the Fractal ADL parser.
     * 
     * @since 2.5
     */
    final public static String CONTEXT_ID_JULIAC = "juliac";
}
